package com.huaxu.service.impl;

import com.huaxu.dao.MonitorDataMapper;
import com.huaxu.entity.DayReportEntity;
import com.huaxu.entity.MonitorDataEntity;
import com.huaxu.quartz.entity.QuartzEntity;
import com.huaxu.quartz.job.MonitorDataReportByDayJob;
import com.huaxu.quartz.job.MonitorDataReportByMonthJob;
import com.huaxu.quartz.job.MonitorDataReportByYearJob;
import com.huaxu.quartz.service.JobAndTriggerService;
import com.huaxu.service.MonitorDataService;
import com.huaxu.util.ByteArrayUtils;
import com.huaxu.util.RedisUtil;
import lombok.Data;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Sort;
import org.springframework.data.mongodb.core.MongoTemplate;
import org.springframework.data.mongodb.core.aggregation.Aggregation;
import org.springframework.data.mongodb.core.aggregation.AggregationOptions;
import org.springframework.data.mongodb.core.aggregation.AggregationResults;
import org.springframework.data.mongodb.core.query.Criteria;
import org.springframework.data.mongodb.core.query.Query;
import org.springframework.data.mongodb.core.query.Update;
import org.springframework.stereotype.Service;

import javax.annotation.Resource;
import java.math.BigDecimal;
import java.time.Duration;
import java.time.LocalDateTime;
import java.util.*;
import java.util.stream.Collectors;

/**
 * @description
 * @auto wangli
 * @data 2020-12-01 11:09
 */
@Service
@Slf4j
public class MonitorDataServiceImpl implements MonitorDataService , InitializingBean {

    @Resource
    private MongoTemplate mongoTemplate;

    @Resource
    private MonitorDataMapper monitorDataMapper;

    @Autowired
    private RedisUtil redisUtil;

    @Autowired
    private JobAndTriggerService jobAndTriggerService;

    //避免同时执行先后顺序造成取数据不准确，时间隔开执行
    @Override
    public void afterPropertiesSet() {
        saveQrtzTask("0 0 */1 * * ? ","日报生成任务","smsWaterMonitorDataReportByDayJob",MonitorDataReportByDayJob.class.getName());
        saveQrtzTask("0 5 0 * * ? ","月报报生成任务","smsWaterMonitorDataReportByMonthJob", MonitorDataReportByMonthJob.class.getName());
        saveQrtzTask("0 10 0 1 * ? ","年报生成任务","smsWaterMonitorDataReportByYearJob", MonitorDataReportByYearJob.class.getName());
    }

    public void saveQrtzTask(String cron, String jobGroup ,String jobName,String JobClassName) {
        // 1,查询需要批量推送的配置项目并构建定时任务
        // 2,若对应定时任务不存在则创建
        QuartzEntity entity = new QuartzEntity();
        entity.setJobGroup(jobGroup);
        entity.setJobName(jobName);
        entity.setDescription(jobName);
        boolean exists = jobAndTriggerService.isExists(entity);
        if(!exists){
            entity.setCronExpression(cron);
            entity.setJobClassName(JobClassName);
            jobAndTriggerService.save(entity);
        }
    }

    @Override
    public MonitorDataEntity save(MonitorDataEntity monitorDataEntity) {
        //缓存数据
        redisUtil.setExpire(("sms_water_"+monitorDataEntity.getDeviceCode()).getBytes(), ByteArrayUtils.objectToBytes(monitorDataEntity).get());
        //把没有数据的属性去掉再保存
        monitorDataEntity.setDataValues(monitorDataEntity.getDataValues().stream().filter(m -> m.getDataValue()!=null).collect(Collectors.toList()));

        return mongoTemplate.save(monitorDataEntity);
    }

    public MonitorDataEntity getDeviceMonitorInfoByDeviceCode(String deviceCode){
        //先取缓存里的数据
        byte[] bytes = redisUtil.get(("sms_water_"+deviceCode).getBytes());
        if(bytes != null && bytes.length>0){
            return (MonitorDataEntity)ByteArrayUtils.bytesToObject(bytes).get();
        }
        return monitorDataMapper.getDeviceMonitorInfoByDeviceCode(deviceCode);
    }

    @Override
    public void remove(Long id) {
        mongoTemplate.remove(id);
    }

    @Override
    public void update(MonitorDataEntity monitorDataEntity) {
        Query query = new Query(Criteria.where("id").is(monitorDataEntity.getId()));

        Update update = new Update();
        update.set("deviceName", monitorDataEntity.getDeviceName());


        mongoTemplate.updateFirst(query, update, MonitorDataEntity.class);
    }

    @Override
    public MonitorDataEntity findById(Long id) {
        Query query = new Query(Criteria.where("id").is(id));
        MonitorDataEntity monitorDataEntity = mongoTemplate.findOne(query, MonitorDataEntity.class);
        return monitorDataEntity;
    }

    @Override
    public List<DayReportEntity> getMonitorDataGroupByHour(LocalDateTime dateTime ){

        if(dateTime == null ){
            return null;
        }
        Criteria criteria = new Criteria();
        criteria.and("year").is(dateTime.getYear());
        criteria.and("month").is(dateTime.getMonthValue());
        criteria.and("day").is(dateTime.getDayOfMonth());
        criteria.and("hour").is(dateTime.getHour());
        AggregationOptions aggregationOptions = AggregationOptions.builder().allowDiskUse(true).build();
        Aggregation agg = Aggregation.newAggregation(
                Aggregation.match(criteria),    //查询条件
                Aggregation.unwind("dataValues"),//将文档中的某一个数组类型字段拆分成多条，每条包含数组中的一个值
                Aggregation.sort(Sort.Direction.ASC, "collectDate"),//在聚合之前对数据进行排序
                Aggregation.group("year","month","day","hour","tenantId","deviceId","deviceName","deviceCode","dataValues.attributeId","dataValues.attributeName" )
                        .min("$dataValues.dataValue").as("minValue")
                        .max("$dataValues.dataValue").as("maxValue")
                        .avg("$dataValues.dataValue").as("avgValue")
//                        .sum("$dataValues.dataValue").as("sumValue")
                        .first("$dataValues.dataValue").as("firstValue")
                        .last("$dataValues.dataValue").as("latestValue")
                        .last("$collectDate").as("collectDate")
                        .count().as("countNum")
        ).withOptions(aggregationOptions);
        AggregationResults<DayReportEntity> durationData =
                mongoTemplate.aggregate(agg, "SMS_MONITOR_DATA", DayReportEntity.class);
        //获取查询结果
        List<DayReportEntity> monthReportEntities = durationData.getMappedResults();
        return monthReportEntities;
    }

    @Override
    public List<DayReportEntity> getMonitorDataReportByHour(){
        //取前一个小时的时间
        LocalDateTime dateTime = LocalDateTime.now().plusHours(-1);

        List<DayReportEntity> hourDatas = getMonitorDataGroupByHour(dateTime);
        List<DayReportEntity> lastHourDatas = getMonitorDataGroupByHour(dateTime.plusHours(-1));

        Map<String,DayReportEntity> lastHourDataMap = lastHourDatas.stream().collect(Collectors.toMap(DayReportEntity::getMapkey, a -> a,(k1, k2)->k1));
        //保存日报表数据
        if(hourDatas.size()>0){
            //计算累计值
            for(DayReportEntity dayReportEntity : hourDatas){
                if(dayReportEntity.getLatestValue()!=null && lastHourDataMap.containsKey(dayReportEntity.getMapkey())){
                    //上一个小时有值
                    if(  lastHourDataMap.get(dayReportEntity.getMapkey()).getLatestValue() != null){
                        dayReportEntity.setSumValue(new BigDecimal(dayReportEntity.getLatestValue().toString()).subtract(new BigDecimal(lastHourDataMap.get(dayReportEntity.getMapkey()).getLatestValue().toString())).doubleValue());
                    }else if(dayReportEntity.getFirstValue() != null){//上一个小时没有值,取本小时的初始值
                        dayReportEntity.setSumValue(new BigDecimal(dayReportEntity.getLatestValue().toString()).subtract(new BigDecimal(dayReportEntity.getFirstValue().toString())).doubleValue());
                    }

                }
            }
            monitorDataMapper.batchInsertDayReport(hourDatas);
        }
        return hourDatas;

//        Map<String,DayReportEntity> hourDataMap = hourDatas.stream().collect(Collectors.toMap(DayReportEntity::getMapkey, a -> a,(k1, k2)->k1));
//        //需要统计的属性值（租户——场景——设备——属性）
//        List<DayReportEntity> dayReportEntities = monitorDataMapper.getReportBaseInfo();
//        //保存日报表数据
//        if(hourDatas.size()>0 && dayReportEntities.size()>0){
//           //填充数据
//
//            dayReportEntities =  dayReportEntities.stream()
//                            .filter(d ->StringUtils.isNotBlank(d.getMapkey()) && hourDataMap.containsKey(d.getMapkey()))
//                            .map(dayReportEntity ->{
//                                DayReportEntity hourData = hourDataMap.get(dayReportEntity.getMapkey());
//                                dayReportEntity.setYear(hourData.getYear());
//                                dayReportEntity.setMonth(hourData.getMonth());
//                                dayReportEntity.setDay(hourData.getDay());
//                                dayReportEntity.setHour(hourData.getHour());
//                                dayReportEntity.setMinValue(hourData.getMinValue());
//                                dayReportEntity.setMaxValue(hourData.getMaxValue());
//                                dayReportEntity.setAvgValue(hourData.getAvgValue());
//                                dayReportEntity.setCollectDate(hourData.getCollectDate());
//                                dayReportEntity.setLatestValue(hourData.getLatestValue());
//                                if(lastHourDataMap.containsKey(dayReportEntity.getMapkey()) && lastHourDataMap.get(dayReportEntity.getMapkey()).getLatestValue() != null){
//                                    dayReportEntity.setSumValue(new BigDecimal(hourDataMap.get(dayReportEntity.getMapkey()).getLatestValue().toString()).subtract(new BigDecimal(lastHourDataMap.get(dayReportEntity.getMapkey()).getLatestValue().toString())).doubleValue());
//                                }
//                                return dayReportEntity;
//                            }).collect(Collectors.toList());
//            monitorDataMapper.batchInsertDayReport(dayReportEntities);
//        }
//        return dayReportEntities;
    }

    @Override
    public void getMonitorDataByDay() {
        //取前一天的时间
        LocalDateTime dateTime = LocalDateTime.now().plusDays(-1);
        //保存日报表数据
        monitorDataMapper.batchInsertMonthReport(dateTime.getYear(),dateTime.getMonthValue(),dateTime.getDayOfMonth());
    }

    @Override
    public void getMonitorDataByMonth() {
        //取前一月的时间
        LocalDateTime dateTime = LocalDateTime.now().plusMonths(-1);
        //保存日报表数据
        monitorDataMapper.batchInsertYearReport(dateTime.getYear(),dateTime.getMonthValue());
    }








}